import pytest
from matrix_operations_cuni.operations import *
from matrix_operations_cuni.core import Matrix


def test_add():
    matrix_one = Matrix([[1, 2, 3], [4, 5, 6]])
    matrix_two = Matrix([[1, 2, 9], [1, 3, 8]])
    addition = add(matrix_one, matrix_two)
    correct_addition = Matrix([[2, 4, 12], [5, 8, 14]])
    assert addition.data == correct_addition.data

    with pytest.raises(ValueError):
        matrix_one = Matrix([[1, 2, 3], [4, 5, 6]])
        matrix_two = Matrix([[1, 2], [1, 3]])
        add(matrix_one, matrix_two)


def test_mul_scalar():
    matrix_one = Matrix([[1, 2, 3], [4, 5, 6]])
    scalar = 0.25
    scalar_mul = mul_scalar(matrix_one, scalar)
    correct_scalar_mul = Matrix([[0.25, 0.5, 0.75], [1, 1.25, 1.5]])
    assert scalar_mul.data == correct_scalar_mul.data


def test_sub():
    matrix_one = Matrix([[1, 2, 3], [4, 5, 6]])
    matrix_two = Matrix([[1, 2, 9], [1, 3, 8]])
    addition = sub(matrix_one, matrix_two)
    correct_addition = Matrix([[0, 0, -6], [3, 2, -2]])
    assert addition.data == correct_addition.data

    with pytest.raises(ValueError):
        matrix_one = Matrix([[1, 2, 3], [4, 5, 6]])
        matrix_two = Matrix([[1, 2], [1, 3]])
        sub(matrix_one, matrix_two)


def test_mat_mul():
    matrix_one = Matrix([[1, 2, 3], [4, 5, 6]])
    matrix_two = Matrix([[1, 2], [9, 1], [1, 3]])
    matrix_mul = mat_mul(matrix_one, matrix_two)
    correct_matrix_mul = Matrix([[22, 13], [55, 31]])
    assert matrix_mul.data == correct_matrix_mul.data

    with pytest.raises(ValueError):
        matrix_one = Matrix([[1, 2, 3], [4, 5, 6]])
        matrix_two = Matrix([[1, 2], [1, 3]])
        mat_mul(matrix_one, matrix_two)



import pytest
from matrix_operations_cuni.core import Matrix


def test__validate_matrix():
    valid_matrix = [[1, 2, 3], [4, 5, 6], [7, 8, 9]]
    assert Matrix._validate_matrix(valid_matrix) is True

    valid_matrix = [[1, 2.6, 3], [4.8, 5, 6], [7, 8.9, 9]]
    assert Matrix._validate_matrix(valid_matrix) is True

    invalid_matrix = [[1, 2], [3, 4, 5]]
    assert Matrix._validate_matrix(invalid_matrix) is False

    invalid_matrix = [[1, 2, "x"], [3, 4, 5]]
    assert Matrix._validate_matrix(invalid_matrix) is False

    invalid_matrix = [[1, "x"], [3, 4, 5]]
    assert Matrix._validate_matrix(invalid_matrix) is False


def test_getitem():
    matrix = Matrix([[1, 2], [3, 4]])
    assert matrix[0, 0] == 1
    assert matrix[1, 1] == 4
    assert matrix[0] == [1, 2]
    with pytest.raises(IndexError):
        _ = matrix[2, 2]
    with pytest.raises(TypeError):
        _ = matrix["invalid"]




import pytest
from matrix_operations_cuni.transformations import Transformations


def test_transpose():
    matrix = Transformations([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    transposed = matrix.transpose()
    expected = Transformations([[1, 4, 7], [2, 5, 8], [3, 6, 9]])
    assert transposed.data == expected.data

    matrix = Transformations([[1]])
    transposed = matrix.transpose()
    expected = Transformations([[1]])
    assert transposed.data == expected.data


def test_rref():
    matrix = Transformations([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    rref_form = matrix.rref()
    correct_rref = Transformations([[1, 0, -1], [0, 1, 2], [0, 0, 0]])
    assert rref_form.data == correct_rref.data

    matrix = Transformations([[1, 2, 0], [0, 5, 6], [7, 0, 9]])
    rref_form = matrix.rref()
    correct_rref = Transformations([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    assert rref_form.data == correct_rref.data

    matrix = Transformations([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    rref_form = matrix.rref()
    correct_rref = Transformations([[0, 0, 0], [0, 0, 0], [0, 0, 0]])
    assert rref_form.data == correct_rref.data

    matrix = Transformations([[1, 0.25, 8], [2, 2, 0], [0, 0, 0]])
    rref_form = matrix.rref()
    rounded_rref = []
    for row in rref_form.data:
        rounded_row = []
        for val in row:
            rounded_row.append(round(val, 2))
        rounded_rref.append(rounded_row)
    correct_rref = Transformations([[1, 0, 10.67], [0, 1, -10.67], [0, 0, 0]])
    assert rounded_rref == correct_rref.data


def test_inverse():
    matrix = Transformations([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    inverse = matrix.inverse()
    correct_inverse = Transformations([[1, 0, 0], [0, 1, 0], [0, 0, 1]])
    assert inverse.data == correct_inverse.data

    matrix = Transformations([[1]])
    inverse = matrix.inverse()
    correct_inverse = Transformations([[1]])
    assert inverse.data == correct_inverse.data

    matrix = Transformations([[2, 2, 3], [1, 3, 4], [1, 1, 1]])
    inverse = matrix.inverse()
    correct_inverse = Transformations(
        [[0.5, -0.5, 0.5], [-1.5, 0.5, 2.5], [1.0, -0.0, -2.0]]
    )
    assert inverse.data == correct_inverse.data

    with pytest.raises(ValueError):
        matrix = Transformations([[2, 2, 3], [1, 3, 4]])
        matrix.inverse()

    with pytest.raises(ValueError):
        matrix = Transformations([[2, 2, 3], [1, 3, 4], [0, 0, 0]])
        matrix.inverse()

    with pytest.raises(ValueError):
        matrix = Transformations([[0]])
        matrix.inverse()


def test_trace():
    matrix = Transformations([[90, 2, 90], [1, 1, 0], [34, 11, 1]])
    trace = matrix.trace()
    correct_trace = 92
    assert trace == correct_trace

    with pytest.raises(ValueError):
        matrix = Transformations([[2, 2, 3], [1, 3, 4]])
        matrix.trace()



class Matrix:
    def __init__(self, data):
        if not self._validate_matrix(data):
            raise ValueError(
                "Invalid matrix data. All rows should be the same size and consist of numbers."
            )
        self.data = data
        self.rows = len(data)
        self.cols = len(data[0])

    @staticmethod
    def _validate_matrix(matrix):
        """
        Validates whether the input follows the matrix structure.
        :param matrix: A nested list of lists.
        :return: True/False based on whether it is a valid representation of a matrix.
        """
        if (
            not isinstance(matrix, list)
            or not all(isinstance(row, list) for row in matrix)
            or not all(
                isinstance(element, (int, float)) for row in matrix for element in row
            )
        ):
            return False
        if (len(matrix) == 0) or (len(set(len(row) for row in matrix)) > 1):
            return False
        return True

    def __repr__(self):
        return str(self.data)

    def __getitem__(self, item: int or tuple):
        """
        Enables to access individual elements of a matrix.
        Return a row by supplying an integer.
        Return an element of a matrix by supplying a tuple
        :param item: which element to return
        :return: chosen element
        """
        if (isinstance(item, tuple)) and (len(item) == 2):
            r_index, c_index = item
            if (0 <= r_index < self.rows) and (0 <= c_index < self.cols):
                return self.data[r_index][c_index]
            raise IndexError("Matrix index out of range")
        elif isinstance(item, int):
            if 0 <= item < self.rows:
                return self.data[item]
            raise IndexError("Matrix index out of range")
        raise TypeError("Matrix indices must be integers or tuples of integers")

    def pretty_print(self):
        """
        Prints the matrix in a simple, readable format.
        Handles -0.0 by converting it to 0.0.
        """
        for r_index in range(self.rows):
            row_values = []
            for c_index in range(self.cols):
                value = self[r_index, c_index]
                if int(value) == value: # fixes -0.0, 2.0, etc.
                    value = int(value)
                if isinstance(value, float):
                    formatted = f"{value:8.4f}" # 4 decimals, 8 chars width
                else:
                    formatted = f"{value:8}"
                row_values.append(formatted)
            print("\t".join(row_values))


if __name__ == "__main__":
    testMatica = Matrix([[1, 2, 3], [3, 3, 4]])
    print(testMatica)
    print(testMatica.rows)
    print(testMatica.cols)
    testMatica.pretty_print()
